/* deliver.c
**
** send a mail to all recipients in a list file
**
*/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/wait.h>

#include "prefs.h"
#include "tools.h"
#include "vxerror.h"
#include "folder.h"
#include "dbif.h"
#include "inline-funcs.h"
#include "mrmacs.h"
#include "tmpfile.h"

#include "userdb.h"
#include "fcheck.h"
#include "hlsock.h"
#include "smtp.h"
#include "rexec.h"
#include "filter/filter.h"

extern char **environ;

/* Diese Routine soll eine Optimierung enthaltenen:
**   Empfnger, die als Domainteil eine der Domains des VxMail-Systems
**   enthalten, sollen direkt zugestellt werden.
*/

static int delivery_level = 0;
static int errflag = 0, de_errs = 0;

/* Diese Variable wird durch das Hauptprogramm belegt. Sie sollte die Lnge der
** die E-Mail enthaltenden temporren Datei enthalten.
*/
size_t msgsize = 0l;

/* Trage die Lnge der E-Mail in die Variable `msgsize' ein. */
void set_msgsize (long sz) { msgsize = sz; }

/* return the number of delivery errors */
int delivery_errs(void) {
    if (prefs.errors == ERRORS_ADMIN && de_errs > 0 && errflag == de_errs) {
	return -de_errs;
    }
    return errflag;
}

/* Sende eine von `vxdeliver' abgewiesene E-Mail an den Default-Administrator
** der betreffenden Domain bzw. `::MAILSYS::' - falls die erste Mglichkeit
** nicht gegeben ist.
*/
void deliver_error (char *rorig, char *sender, char *em_addr,
		    FILE *tfp, char *message, int err);

/* Verschicke eine E-Mail an alle in einer Mailingliste stehenden Empfnger. */
static void deliver_list (char *rorig, char *sender, char *addr,
			  FILE *tfp, char *add_domain, int err);

/* Schreibe eine E-Mail in die Standardeingabe eines zu diesem Zweck von
** `vxdeliver' gestarteten Programmes.
*/
static void deliver_prog (char *rorig, char *sender, char *cmdline,
			  FILE *tfp, int err);

/* Hnge eine E-Mail an eine Datei (Textdatei oder xFile) an. */
static void deliver_file (char *rorig, char *sender, char *outfile,
			  FILE *tfp, int err);

/* send a message to a (potentially non-local) user */
static void deliver_nonlocal (char *rorig, char *sender, char *addr,
			      FILE *tfp, int err);

static void deliver_smtp (char *rorig, char *sender, char *recip_host,
			  char *recipient, FILE *tfp, int err);

static void deliver_bsmtp (char *sender, char *_batchfile,
			   char *recipient, FILE *tfp,
			   size_t ml, off_t fl, int err);

/* berlies eine eventuell in der die E-Mail enthaltenden temporren Datei
** vorkommende `From ...'-Zeile.
*/
static off_t skip_From (FILE *tfp);

/* create three parameters from either the sender or the recipient address */
static void addrpars (char *parm, int what, char *vars[]);

/* Liefere eine Meldung an einen lokalen Empfnger aus */
void deliver_local (char *rorig, char *sender, char *em_addr, FILE *tfp, int err)
{
    int rtype, eolf = 0, filter_res, ix;
    char *folder = NULL, line[1024], *p, *flt, *vars[11], dtsstrg[60];
    time_t dts;
    FILE *ffp, *tfpc = (FILE *) NULL;
    xFile tgt;
    size_t ml; off_t fl;
    filter_t filter;
    (void) time(&dts);
    (void) strftime(dtsstrg, 60, "%a, %d %b %Y %H:%M:%S %z", localtime(&dts));
    if (rorig && !strcmp (rorig, em_addr)) {
	vxlog (0, "delivery depth level exceeded!\n");
	deliver_error (NULL, sender, rorig, tfp,
		       "recursive mail target found!\n", err);
	return;
    }
    if (delivery_level > prefs.maxlevel) {
	deliver_error (rorig, sender, em_addr, tfp,
		       "maximum delivery level exceeded; you should check "
		       "if one of your\n"
		       "`list link' entries is a recursive reference "
		       "(directly or indirectly)\n"
		       "to itself.\n", err);
	return;
    }
    ++delivery_level;
    /* check if the corresponding mail filter rejects the e-mail */
    if (!fcheck (em_addr, tfp)) {
	if (prefs.rejverb) {
	    vxlog (0, "mail rejected, because it matches the filter "
		      "conditions!\n");
	    deliver_error (NULL, sender, rorig, tfp,
			   "mail rejected by domain filter", err);
	}
	--delivery_level;
	return;
    }
    /* try to find the recipient */
    ifnull (folder = user_folder (em_addr, &ml, &fl, &rtype)) {
	vxlog (1, "`%s' not found - ", em_addr);
	folder = user_rewrite (em_addr, &ml, &fl, &rtype);
    }
    vxlog (5, "folder = `%s'%s\n", (rtype == RT_LLINK ? &folder[1] : folder),
				   (rtype == RT_LLINK ? "..." : ""));
    /* if the recipient does not exist, send an error message to the
    ** corresponding default domain administrator (or to ::MAILSYS::)
    */
    ifnull (folder) {
	p = gconc (__FILE__, __LINE__, "recipient `", em_addr, "' not found",
				       " in the VxMail database.\n", NULL);
	vxlog (0, "attempt to write to folder failed!\n");
	deliver_error (rorig, sender, em_addr, tfp, p, err);
	gfree (p);
	--delivery_level;
	return;
    }
    if (rtype == RT_BSMTP) {
	char *recipient = (rorig != NULL ? rorig : em_addr);
	/* (Syntax: `/var/spool/mail/vx/bsmtp/relay' (Beispiel))
	** Die E-Mail soll in einem VxMail-Folder als halbfertige SMTP-
	** Session abgespeichert werden und per SMTP spter an eine
	** dynamische Zieladresse gesendet werden - getriggert durch
	** eine POP3-Anfrage auf diesen Benutzereintrag; dazu mu ein
	** Pawort beim Anlegen dieses Benutzereintrages in der Datenbank
	** mit angegeben werden. Auch dieser Eintrag ergibt nur dann einen
	** wirklichen Sinn, wenn mehrere Empfnger (oder eine ganze Domain)
	** auf ihn verweisen.
	*/
	deliver_bsmtp (sender, folder, recipient, tfp, ml, fl, err);
	gfree (folder);
	--delivery_level;
	return;
    } else if (rtype == RT_LLINK) {
	char *recipient = (rorig != NULL ? rorig : em_addr), *fld = folder;
	int mx = ((int) *fld++) + 1, ix;
	size_t sl = 0;

	for (ix = 0; ix < mx; fld += sl, ++ix) {
	    sl = strlen (fld) + 1;
	    if (*fld == '|') {
		/* (Syntax: `|/usr/local/lib/mlsender' (Beispiel))
		** Die bermittlung der E-Mail soll durch ein Programm
		** geschehen. blicherweise ist dies ein Programm zur
		** Verwaltung von Mailinglisten.
		*/
		deliver_prog (recipient, sender, fld + 1, tfp, err);
	    } else if (*fld == '>') {
		/* (Syntax: `>/var/lists/Data/liste1.txt' oder
		**          `>>/var/lists/Data/liste1.txt' oder
		**          `>:/var/lists/Data/liste1.txt' (Beispiel))
		** Die E-Mail soll an eine Textdatei oder einen VxMail-Folder
		** angehngt werden - interessant im Zusammenhang mit der
		** Archivierung von E-Mails an Mailinglisten.
		*/
		deliver_file (recipient, sender, fld + 1, tfp, err);
	    } else if (*fld == '/') {
		/* (Syntax: `/var/vxmail/liste1' (Beispiel))
		** Die E-Mail soll an alle in der angegebenen Datei stehenden
		** Empfnger weitergereicht werden; es handelt sich hierbei um
		** einen einfachen Mechanismus fr Mailinglisten und Aliases mit
		** mehr als einem Adressaten.
		** Die Mailingliste mu in diesem Fall als absoluter Pfadname
		** angegeben werden.
		*/
		char *add_domain = NULL;
		unlessnull (p = strchr (fld, ':')) {
		    *p++ = '\0';
		    if (*p) {
			add_domain = gdupl (__FILE__, __LINE__, p);
		    } else unlessnull (p = strchr (recipient, '@')) {
			add_domain = gdupl (__FILE__, __LINE__, ++p);
		    }
		} else unlessnull (p = strchr (recipient, '@')) {
		    add_domain = gdupl (__FILE__, __LINE__, ++p);
		}
		deliver_list (recipient, sender, fld, tfp, add_domain, err);
		unlessnull(add_domain) { gfree(add_domain); add_domain = NULL; }
	    } else unlessnull (strchr (fld, '@')) {
		/* (Syntax: `fbj@blinx.de' (Beispiel))
		** Ein (mglicherweise nicht-lokaler) Empfnger ist das Ziel
		** der E-Mail. Dies eignet sich fr die bertragung von E-Mails
		** an einzelne Empfnger auf einem anderen Rechner.
		*/
		deliver_nonlocal (recipient, sender, fld, tfp, err);
	    } else if (!strncmp (fld, "smtp:", 5)) {
		/* (Syntax: `smtp:mail.zweitehand.de' oder
		** `smtp:[194.115.191.214]' (Beispiel))
		** Eine SMTP-bertragung an den durch <hostname> oder <ip-addr>
		** angegebenen Host soll stattfinden. Dies eignet sich fr eine
		** Direkbertragung an eine nicht auf dem gleichen Rechner
		** liegende Domain - bzw. jeden in dieser Domain stehenden
		** Benutzer.
		** Das ganze ergibt natrlich nur dann einen Sinn, wenn mehrere
		** E-Mail-Empfnger auf diesen Eintrag verweisen
		** (USER "<OTHER>"), da ansonsten der Empfnger u.U. nicht
		** korrekt ermittelt werden kann; auerdem mu der empfangende
		** SMTP-Server die Domain als seine eigene (lokale Domain)
		** werten (oder evtl. als `virtual'?:).
		** (Z. Zt. noch nicht implementiert.)
		*/
		deliver_smtp (rorig, sender, fld + 5, em_addr, tfp, err);
	    } else unlessnull (p = strchr (&fld[1], ':')) {
		/* (Syntax: `VxMail-Liste:', `xyz.liste:xyz.de' (Beispiele))
		** Die E-Mail soll an alle in der angegebenen Datei stehenden
		** Empfnger weitergereicht werden; es handelt sich hierbei um
		** einen einfachen Mechanismus fr Mailinglisten und Aliases mit
		** mehr als einem Adressaten.
		** Die Mailingliste mu in diesem Fall in dem durch die VxMail-
		** Konfiguration (`listdir /var/VxMail/lists') spezifizierten
		** Verzeichnis stehen.
		*/
		/* NEW in VxMail-1.48:
		** extract the domain postfix from a listfile entry and use
		** it for list entries which don't have a domain part.
		** If the domain postfix is empty, then try the domain of the
		** user entry the message was originally sent to.
		*/
		char *newpath, *add_domain = NULL;
		*p++ = '\0';
		if (*p) {
		    add_domain = gdupl (__FILE__, __LINE__, p);
		} else unlessnull (p = strchr (recipient, '@')) {
		    add_domain = gdupl (__FILE__, __LINE__, ++p);
		}
		newpath = gconc(__FILE__, __LINE__,
				prefs.listdir, "/", fld, NULL);
		deliver_list (recipient, sender, newpath, tfp, add_domain, err);
		gfree (newpath);
		unlessnull(add_domain) { gfree(add_domain); add_domain = NULL; }
	    } else {
		/* NEW in VxMail-1.48:
		** A LLTO entry which does not match one of the categories above
		** is a link to a user entry without domain part; in this case
		** the domain of the original user should be appended.
		*/
		char *new_user = NULL;
		unlessnull (p = strchr (recipient, '@')) {
		    new_user = gconc (__FILE__, __LINE__, fld, p, NULL);
		    deliver_nonlocal (recipient, sender, new_user, tfp, err);
		    gfree (new_user);
		} else {
		    deliver_error (recipient, sender, fld, tfp,
				   "The target of this LLTO entry does "
				   "not have a domain part!", err);
		}
	    }
	}
	gfree (folder);
	--delivery_level;
	return;
    }
    /* Liefere die E-Mail ganz normal aus */
    if (!(tgt = flOpen (folder, "wx", prefs.fperm))) {
	int err = errno;
	if (!prefs.syswithidx && rtype == RT_NSUSER) xNoIndex (folder);
	errno = err;
	vxerror_at (EX_CANTCREAT, errno, __FILE__, __LINE__ - 4,
		    "writing folder for user `%s' failed", em_addr);
    }
    vxlog(5, "maximum message size = %ld\n", ml);
    if (ml && msgsize > ml) {
	/* Die E-Mail mu abgewiesen werden, weil die ihre Lnge die fr den
	** betreffenden Benutzer zugelassene maximale Lnge von E-Mails
	** berschreiten wrde.
	*/
	xClose (tgt);
/*	if (!prefs.syswithidx && rtype == RT_NSUSER) xNoIndex (folder);*/
	vxlog (0, "maximum message size exceeded!\n");
	deliver_error (rorig, sender, em_addr, tfp,
		       "your message exceeds the allowed size", err);
	--delivery_level;
	return;
    }
    vxlog(5, "maximum folder size = %ld\n", fl);
    if (fl && xSize(tgt)+(off_t)msgsize > fl) {
	/* Die E-Mail mu abgwiesen werden, weil die Gesamtlnge des
	** Folders mit dieser E-Mail die Maximallnge berschreiten wrde.
	*/
	xClose (tgt);
/*	if (!prefs.syswithidx && rtype == RT_NSUSER) xNoIndex (folder);*/
	vxlog (0, "maximum folder size exceeded!\n");
	deliver_error (rorig, sender, em_addr, tfp,
		       "the mailbox is full", err);
	--delivery_level;
	return;
    }
/*
    xChown (tgt, geteuid (), getgid ());
    xChmod (tgt, prefs.fperm);
*/
/*##CHANGE: 08/07/2001*/
    /* The filter actions MUST NOT modify the original mail text, because
    ** this mail text may also be delivered to other recipients. As a result,
    ** the filter file is copied...
    */
    if (fvxtrans (tfp, &tfpc)) {
	vxlog (0, "FATAL! Attempt to copy file handle failed - %s.\n",
		  strerror (errno));
	vxerror_at (EX_OSERR, errno, __FILE__, __LINE__ - 3,
		    "FATAL! Attempt to copy file handle failed");
    }
    flt = gconc (__FILE__, __LINE__, folder, "#flt", NULL);
    ifnull (ffp = fopen (flt, "r")) {
	vxlog (1, "no filter found - going on normal (error: %s)\n",
		  strerror(errno));
	gfree (flt);
	goto NOFLT;
    }
    vxlog (1, "using mail filter `%s'.\n", flt);
    /* create the parameters for the parameter substitution in `fparser'. */
    /* SENDER, SUSER, SDOMAIN */
    addrpars (sender, 0, vars);
    /* RORIG, RUSER, RDOMAIN (original recipient) */
    addrpars ((isNULL(rorig) ? em_addr : rorig), 1, &vars[3]);
    /* TARGET, TUSER, TDOMAIN (current recipient) */
    addrpars (em_addr, 2, &vars[6]);
    /* DTS (the delivery timestamp in RFC822 format) */
    vars[9] = gconc(__FILE__, __LINE__, "DTS=", dtsstrg, NULL);
    /* conclude this by an end-marker. */
    vars[10] = NULL;
    /* call the filter parser with the opened filter file and the just
    ** generated parameters.
    */
    filter = fparser (ffp, vars, sender);
    /* free the memory allocated for the parameters when the parser completed
    ** it's operation.
    */
    for (ix = 0; ix < 9; ++ix) { gfree (vars[ix]); vars[ix] = NULL; }
    /* Close the filter file and free the buffer for it's name. */
    fclose (ffp); gfree (flt);
    /* Continue with the normal delivery if the filter parsing failed. */
    ifnull (filter) { goto NOFLT; }
    /* Execute the filter interpreter on the e-mail. */
    filter_res = finterp (filter, tfpc, tgt, err);
    /* Free the filter memory when `finterp' completed it's operations. */
    filter_free (filter);
    /* If `finterp' either failed or executed no `Mailbox' or `Drop' action,
    ** then continue as if no filter as been applied to the e-mail.
    */
    if (filter_res != 0) { goto NOFLT; }
    vxlog (1, "... done (by filter).\n", filter_res);
    /* Otherwise, the filter completed the delivery operation. */
    goto DONE;
NOFLT:
/*##END OF CHANGE*/
    /* deliver the e-mail normally. */
    rewind(tfpc);
    vxlog (1, "delivering message to `%s' ", em_addr);
    while (fgets (line, 1024, tfpc)) {
	eolf = cuteol (line); xPuts (line, tgt);
	if (eolf) { xPuts ("\r\n", tgt); }
	/*vxlog (5, ".");*/
    }
    if (!eolf) { xPuts ("\r\n", tgt); }
    vxlog (1, "... done.\n");
DONE:
    fclose (tfpc);
    flSetNextUID (tgt);
    xClose (tgt);
    if (!prefs.syswithidx && rtype == RT_NSUSER) { xNoIndex (folder); }
    rewind (tfp);
    gfree (folder);
    --delivery_level;
}

static void addrpars (char *parm, int what, char *vars[])
{
    char *vu, *vd;
    char *pars[3];
    pars[0] = (what == 0 ? "SENDER"  : what == 1 ? "RECIPIENT" : "TARGET");
    pars[1] = (what == 0 ? "SDOMAIN" : what == 1 ? "RDOMAIN"   : "TDOMAIN");
    pars[2] = (what == 0 ? "SUSER"   : what == 1 ? "RUSER"     : "TUSER");
    vars[0] = gconc (__FILE__, __LINE__, pars[0], "=", parm, NULL);
    vd = strrchr (parm, '@');
    vars[1] = gconc (__FILE__, __LINE__,
		     pars[1], "=", isNULL(vd) ? vd : &vd[1], NULL);
    ifnull (vd) {
	vars[2] = gconc (__FILE__, __LINE__, pars[2], "=", parm, NULL);
    } else {
	vu = galloc (__FILE__, __LINE__, (vd - parm) + 1);
	memcpy (vu, parm, (vd - parm));
	vu[vd - parm] = '\0';
	vars[2] = gconc (__FILE__, __LINE__, pars[2], "=", vu, NULL);
	gfree (vu);
    }
}

/* Liefere eine von `vxdeliver' abgewiesene E-Mail an den jeweiligen
** Domain-Administrator bzw. `::MAILSYS::' aus.
*/
void deliver_error (char *rorig, char *sender, char *em_addr,
		    FILE *tfp, char *message, int err)
{
    int lasteol, waseol, ctf = 0;
    char *admname, *domain, *admin, line[1024], *adms, *ct1 = NULL, *eol,
	 boundary[60];
    FILE *efp;
    time_t tim;
    ++errflag;
    vxlog (1, "error delivery mode ... ");
    if (prefs.errors == ERRORS_SYSTEM) { vxlog (1, "system.\n"); return; }
    if (err) {
	vxlog (1, "already in error delivering mode - switching to system"
		  "mode.\n");
	++de_errs;
	return;
    }
    vxlog (1, "%s.\n", prefs.errors == ERRORS_ADMIN ? "admin" : "admin+system");
    admname = defadmin ();
    if (strcmp (em_addr, "::MAILSYS::")) {
	ifnull (domain = strchr (em_addr, '@')) {
	    vxerror_at (EX_SOFTWARE, 0, __FILE__, __LINE__ - 2,
			"e-mail address is invalid (no domain)!\n");
	}
	admin = gconc (__FILE__, __LINE__, admname, domain, NULL);
    } else {
	admin = gdupl (__FILE__, __LINE__, "::MAILSYS::");
    }
    ifnull (efp = fvxtmp ()) {
	vxerror_at (EX_OSERR, errno, __FILE__, __LINE__ - 1,
		    "attempt to write temporary file failed");
    }
    (void) time (&tim);
    strftime (line, 1024, "%a, %d %b %Y %H:%M:%S %z", localtime (&tim));
    fprintf (efp, "From VxMail@%s %s\r\n", mailhost(), line);
    adms = gvarins (__FILE__, __LINE__,
		    prefs.error_sender, "MAILHOST", mailhost ());
    fprintf (efp, "From: %s\r\n", adms);
    fprintf (efp, "To: %s\r\n", em_addr);
    fprintf (efp, "Subject: mail delivery to <%s> failed\r\n", em_addr);
    fprintf (efp, "Date: %s\r\n", line);
    fprintf (efp, "MIME-Version: 1.0\r\n");
    fprintf (efp, "Content-Type: multipart/report;\r\n"
		  "  report-type=delivery-status;\r\n");
    sprintf (boundary, "_===%08x.%08x===_", getpid(), rand());
    fprintf (efp, "  boundary=\"%s\"\r\n\r\n", boundary);
    fprintf (efp, "\r\n\r\n--%s\r\nContent-Type: text/plain\r\n", boundary);
    fprintf (efp, "Subject: \r\n\r\n%s\r\n",
		  (isNULL(message) ? "unknown error" : message));
    fprintf (efp, "\r\n--%s\r\n", boundary);
    fprintf (efp, "Content-Type: text/rfc822-headers\r\n"
		  "Subject: message headers\r\n\r\n");
    rewind (tfp);
    /* Skip the `From '-line... */
    unlessnull (fgets (line, 1024, tfp)) {
	if (strncmp (line, "From ", 5)) {
	    rewind (tfp); 
	} else {
	    lasteol = cuteol (line);
	    while (!lasteol) {
		ifnull (fgets (line, 1024, tfp)) { break; }
		lasteol = cuteol (line);
	    }
	}
    }
    /* print the headers of the original e-mail to the file... */
    lasteol = 1; ctf = 0; ct1 = NULL;
    while (noNULL(fgets (line, 1024, tfp))) {
	waseol = cuteol (line);
	if (waseol && *line == '\0') { break; }
	fputs (line, efp);
	if (waseol) { fputs ("\r\n", efp); }
	if (lasteol) {
	    if (!lccmp (line, "Content-Type: ", 14)
	    ||  !lccmp (line, "Content-Encoding: ", 18)) {
		ctf = 1;
		eol = (waseol ? "\r\n" : "");
		gappend (__FILE__, __LINE__, &ct1, line, eol, NULL);
	    } else if (ctf) {
		if (isws (*line)) {
		    eol = (waseol ? "\r\n" : "");
		    gappend (__FILE__, __LINE__, &ct1, line, eol, NULL);
		} else {
		    ctf = 0;
		}
	    }
	} else if (ctf) {
	    eol = (waseol ? "\r\n" : "");
	    gappend (__FILE__, __LINE__, &ct1, line, eol, NULL);
	}
	lasteol = waseol;
    }
    /* print the message body of the original e-mail to the file... */
    fprintf (efp, "\r\n--%s\r\n", boundary);
    if (noNULL(ct1) && prefs.biirm) {
	/* Insert the mime headers from the original message - but only
	** if the body is to be included into the error mail.
	*/
	fprintf (efp, "%s", ct1);
    } else {
	/* If there are no mime headers or the body of the original message
	** is not included into the error mail, the content type is
	** `text/plain' and the encoding 8bit (for an un-encoded message).
	*/
	fprintf (efp, "Content-Type: text/plain; charset=iso-8859-1\r\n"
		      "Content-Transfer-Encoding: 8bit\r\n");
    }
    /* The subject for this part of the error mail is _always_
    ** `message body'.
    */
    fputs ("Subject: message body\r\n\r\n", efp);
    if (prefs.biirm) {
	while (noNULL (fgets (line, 1024, tfp))) {
	    lasteol = cuteol (line);
	    fputs (line, efp);
	    if (lasteol) { fputs ("\r\n", efp); }
	}
    } else {
	fputs ("[message body not included]\r\n", efp);
    }
    /* Close this e-mail by adding the end message bonudary... */
    fprintf (efp, "\r\n--%s--\r\n\r\n", boundary);
    fflush (efp);
    rewind (efp);
    /* deliver the new message locally. */
    deliver_local (rorig, adms, admin, efp, 1);
    /* free the allocated resources and close the temporary file which contains
    ** the error message.
    */
    gfree (adms); gfree (admin); fclose (efp); rewind (tfp);
}

/* Liefere E-Mails an Empfnger, die in einer Listendatei stehen, oder aber
** an ein Programm oder eine Datei oder einen lokalen oder nicht-lokalen
** Benutzer aus.
*/
static void deliver_list (char *rorig, char *sender, char *addr,
			  FILE *tfp, char *add_domain, int err)
{
    char line[1024], rl[1024],
	 *newlist = NULL, *new_addr = NULL, *new_dom = NULL;
    FILE *lf;
    int skipeol;
    ifnull (lf = fopen (addr, "r")) {
	vxlog (0, "opening mailing list `%s' failed!\n", addr);
	deliver_error (rorig, sender, rorig, tfp, strerror (errno), err);
	return;
    }
    while (noNULL(fgets (line, 1024, lf))) {
	skipeol = !cuteol (line);
	while (skipeol) {
	    if (!fgets (rl, 1024, lf)) { break; }
	    skipeol = !cuteol (rl);
	}
	if (!*line || *line == '#') continue;
	if (*line == '|') {
	    deliver_prog (rorig, sender, line + 1, tfp, err);
	} else if (*line == '>') {
	    deliver_file (rorig, sender, line + 1, tfp, err);
	} else if (*line != '/' && strchr (line, '@')) {
	    deliver_nonlocal (rorig, sender, line, tfp, err);
	} else {
	    /* NEW in VxMail-1.46:
	    ** In order to make this working with `majordomo' mailing lists
	    ** plus archiving, i should allow at least one recursion step
	    ** here. But this can only work correctly if the list file does
	    ** not refer itself.
	    */
	    if (*line == '/') {
		/* NEW in VxMail-1.48:
		** If the filename ends with a domain name (e.g.
		** `xyz.list:isv-gmbh.de', then each entry in this list which
		** does not have a domain part gets this domain appended.
		*/
		unlessnull (new_addr = strchr (&line[1], ':')) {
		    *new_addr++ = '\0';
		    if (*new_addr) {
			new_dom = gdupl (__FILE__, __LINE__, new_addr);
		    }
		}
		newlist = gdupl(__FILE__,__LINE__, line);
	    } else unlessnull (new_addr = strchr (&line[1], ':')) {
		*new_addr++ = '\0';
		if (*new_addr) {
		    new_dom = gdupl (__FILE__, __LINE__, new_addr);
		}
		newlist = gconc(__FILE__,__LINE__, prefs.listdir,"/",line,NULL);
	    } else {
		/* NEW in VxMail-1.48:
		** Each entry without a domain part is either an error
		** or - if a value for `add_domain' is specified - a user
		** which has the domain of the corresponding mailing list
		** to be appended.
		*/
		ifnull (add_domain) {
		    deliver_error (rorig, sender, rorig, tfp,
				   "the list file entry has no domain part "
				   "and there was no domain\n"
				   "specified with the list file containing "
				   "this entry...\n", err);
		} else {
		    new_addr = gconc (__FILE__, __LINE__,
				      line, "@", add_domain, NULL);
		    deliver_nonlocal (rorig, sender, new_addr, tfp, err);
		    gfree (new_addr); new_addr = NULL;
		}
		continue;
	    }

	    if (!strcmp (addr, line)) {
		/* list file and target list are the same. This would lead to
		** an endless recursion, so it's forbidden here!
		*/
		vxlog (0, "recipient mailing list `%s' refers to itself!\n",
		       addr);
		deliver_error (rorig, sender, rorig, tfp,
			       "there is definitely an entry in the recipient "
			       "mailing list\n"
			       "referring to the mailing list itself!\n"
			       "This is ILLEGAL!\n", err);
	    } else if (delivery_level > prefs.maxlevel) {
		/* indirect recursion can not be detected properly; the method
		** i chose for doing this is a `level' counter. If this
		** counter exceeds a specified limit (see: `/etc/vxmail.conf'),
		** i assume that an indirectly recursive entry was found.
		** In this case, the message must be rejected and an error
		** be sent to the message's originator instead.
		*/
		deliver_error (rorig, sender, rorig, tfp,
			       "maximum delivery level exceeded; you should "
			       "check if one of your\n"
			       "`list link' entries is a recursive reference "
			       "(directly or indirectly)\n"
			       "to itself.\n", err);
	    } else {
		/* Increase the delivery level before calling yourself */
		/* NEW in VxMail-1.48:
		** if an entry in the list doesn't have a domain part,
		** either `new_dom' (if defined) or `add_domain' will
		** be appended as domain part.
		*/
		++delivery_level;
		deliver_list (rorig, sender, newlist, tfp,
			      (isNULL(new_dom)? add_domain : new_dom), err);
		/* - and decrease it afterwards. */
		--delivery_level;
	    }
	    /* the temporarily allocated memory (for the list file name) must
	    ** be free'd, before we leave this section.
	    */
	    unlessnull (new_dom) { gfree (new_dom); new_dom = NULL; }
	    unlessnull (newlist) { gfree (newlist); newlist = NULL; }
	}
    }
    fclose (lf);
    return;
}

static char *shellpath = "/bin/sh";
static char *shell[] = { "sh", "-c", NULL, NULL };

/* Liefere eine E-Mail an ein Programm aus. */
static void deliver_prog (char *rorig, char *sender, char *cmdline,
			  FILE *tfp, int err)
{
    int fd, pst;
    char *newcmd, *newcmd1, *msg, *rdom, *ruser, *sdom, *suser;
    pid_t shpid;
    uid_t tuid;
    gid_t tgid;
    vxlog (2, "Sending message to `stdin' of a command.\n");
    ifnull (newcmd = valid_prog (cmdline, &tuid, &tgid, 0)) {
	msg = gconc (__FILE__, __LINE__,
		     "recipient program "
		     "cannot be trusted enough\nfor beeing executed!\n"
		     "The original command was `", cmdline, "'.",
		     NULL);
	deliver_error (rorig, sender, rorig, tfp, msg, err);
	gfree (msg);
	vxlog (0, "failed - program is not in the `trustee' list.\n");
	return;
    }
    if (getuid () != 0 && (geteuid () != tuid || getgid () != tgid)) {
	msg = gconc (__FILE__, __LINE__,
		     "' cannot execute the recipient program with the "
		     "specified uid/gid.\n",
		     NULL);
	deliver_error (rorig, sender, rorig, tfp, msg, err);
	gfree (msg);
	vxlog (0, "failed - program can't be executed with UID = %d, "
		  "GID = %d.\n", tuid, tgid);
	return;
    }
    rdom = strchr (rorig, '@');
    ifnull (rdom) {
	ruser = gdupl (__FILE__, __LINE__, rorig);
	rdom = mailhost();
    } else {
	++rdom;
	ruser = galloc (__FILE__, __LINE__, rdom - rorig);
	memcpy (ruser, rorig, rdom - rorig - 1);
	ruser[rdom - rorig - 1] = '\0';
    }
    sdom = strchr (sender, '@');
    ifnull (sdom) {
	suser = gdupl (__FILE__, __LINE__, sender);
	sdom = "no-domain";
    } else {
	++sdom;
	suser = galloc (__FILE__, __LINE__, sdom - sender);
	memcpy (suser, sender, sdom - sender - 1);
	suser[sdom - sender - 1] = '\0';
    }
    
    newcmd1 = gvarins (__FILE__, __LINE__, newcmd, "RECIPIENT", rorig);
    gfree (newcmd); newcmd = NULL;
    newcmd = gvarins (__FILE__, __LINE__, newcmd1, "SENDER", sender);
    gfree (newcmd1); newcmd1 = NULL;
    newcmd1 = gvarins (__FILE__, __LINE__, newcmd, "RUSER", ruser);
    gfree (newcmd); newcmd = NULL;
    newcmd = gvarins (__FILE__, __LINE__, newcmd1, "RDOMAIN", rdom);
    gfree (newcmd1); newcmd1 = NULL;
    newcmd1 = gvarins (__FILE__, __LINE__, newcmd, "SUSER", suser);
    gfree (newcmd); newcmd = NULL;
    newcmd = gvarins (__FILE__, __LINE__, newcmd1, "SDOMAIN", sdom);
    gfree (newcmd1); newcmd1 = NULL;
    gfree (ruser); gfree (suser);
    vxlog (3, "%s <?\n", newcmd);
    rewind (tfp);
    fd = fileno (tfp);
    switch ((shpid = fork ())) {
	case -1:
	    vxerror_at (EX_OSERR, errno, __FILE__, __LINE__ - 2, "(fork)");
	case 0:
	    dup2 (fd, 0);
	    fclose (tfp);
	    shell[2] = newcmd;
	    setreuid (getuid (), getuid ());
	    setgid (tgid); setuid (tuid);
	    execve (shellpath, shell, environ);
	    vxerror_at (EX_OSERR, errno, __FILE__, __LINE__ - 1, "(execve)");
	default:
	    if (waitpid (shpid, &pst, 0) < 0) { return; }
	    if ((WIFEXITED(pst) && WEXITSTATUS(pst)) || WIFSIGNALED(pst)) {
		unlessnull (newcmd) { gfree (newcmd); newcmd = NULL; }
		return;
	    }
    }
    unlessnull (newcmd) { gfree (newcmd); newcmd = NULL; }
    vxlog (2, "Sending message to command succeeded.\n");
    return;
}

/* Hnge eine E-Mail an eine Datei an:
** Wurde `>:dateiname' angegeben, dann wird die E-Mail an eine Datei vom Typ
** `xFile' (Folder mit Index) angehngt; wurde '>dateiname' oder `>>dateiname'
** angegeben, dann wird die E-Mail an eine normale Textdatei angehngt.
*/
static void deliver_file (char *rorig, char *sender, char *outfile,
			  FILE *tfp, int err)
{
    int is_xFile = (*outfile == ':');
    xFile out;
    char *msg = NULL, line[1024];
    if (*outfile == ':' || *outfile == '>') { ++outfile; }
    if (!(out = flOpen (outfile, "wx", prefs.fperm))) {
	msg = gconc (__FILE__, __LINE__,
		     "attempt to append the message to the user's folder "
		     "failed - ", strerror (errno), ".\n", NULL);
	vxlog(0, "attempt to append message to folder `%s' failed - %s.\n",
		 outfile, strerror(errno));
	deliver_error (rorig, sender, rorig, tfp, msg, err);
	gfree (msg);
	return;
    }
    xChown (out, geteuid (), getgid ());
    xChmod (out, prefs.fperm);
    rewind (tfp);
    while (noNULL(fgets (line, 1024, tfp))) { xPuts (line, out); }
    rewind (tfp);
    flSetNextUID (out);
    xClose (out); if (!is_xFile) { xNoIndex (outfile); }
    return;
}

/* Liefere eine E-Mail an einen (potentiell nicht-lokalen) Empfnger aus;
** `vxdeliver' erkennt hierbei, ob ein Empfnger lokal ist daran, da die
** entsprechende Domain in der VxMail-Datenbank eingetragen ist; steht sie
** dort nicht, dann versucht `vxdeliver', die E-Mail an das in der
** VxMail-Konfiguration unter `mailcmd' angegebene Programm zu bergeben.
** In der hinter `mailcmd' angegebenen Kommandozeile knnen die Parameter
** $(SENDER), $(RECIPIENT) und $(SUBJECT) nebst den fr das jeweilige Programm
** sonst noch bentigten Argumenten bergeben werden; erstere werden durch
** die Absenderadresse, die Empfngeradresse und das in der E-Mail-Datei
** enthaltene `subject' ersetzt; von der E-Mail wird nur der Rumpf an das
** gestartete Programm bergeben. Da das Programm als `/bin/sh'-Kommando
** gestartet wird, mssen Absenderadresse, Empfngeradresse und `subject'
** qequotet werden; dazu werden alle im `subject' enthaltenen `"' in `\"'
** verwandelt; da weder Absenderadresse noch Empfngeradresse dieses Zeichen
** enthalten drfen, geschieht dir Umwandlung dort nicht.
** Anschlieend werden alle drei Angaben in Paare von `"' eingeschlossen und
** an der durch das betreffende Symbol festgelegten Stelle eingesetzt.
*/

static char *msgs[] = {
    "internal/operating system error",
    "smtp connection failed",
    "HELO command failed",
    "MAIL command failed",
    "RCPT command failed",
    "DATA command failed",
    "transmitting mail body failed",
    "QUIT command failed"
};

/* deliver_smtp (rorig, sender, recip_host, recipient, tfp, err)
**    sendet die E-Mail direkt an den Empfnger `recipient' auf dem Server
**    `recip_host' per SMTP weiter. Diese Variante der Auslieferung ist
**    fr Standverbindungen zu anderen Hosts gedacht, die die entsprechende
**    E-Mail-Domain lokal (oder virtuell) behandeln knnen.
*/
static void deliver_smtp (char *rorig,
			  char *sender,
			  char *recip_host,
			  char *recipient,
			  FILE *tfp,
			  int err)
{
    /* Eine Verbindung mit dem Zielserver soll sowohl per Namen als auch per
    ** Adresse erfolgen knnen, je nachdem, ob `recip_host' ohne oder mit `['
    ** beginnt, oder nicht.
    */
    char line[1024], *swd = NULL;
    int ev, portnum;
    HLSock hlc;
    off_t fp = skip_From (tfp);
    portnum = LLPort ("smtp", "tcp");
    if (portnum < 0) {
	sprintf (line, "mailcmd internal: service ==> portnum translation "
		       "failed!\n");
	vxlog (0, "initialising SMTP connection failed - unknown service!\n");
	deliver_error (rorig, sender, rorig, tfp, line, err);
	return;
    }
    /* Verbindung zum `relay host' herstellen */
    ifnull (hlc = HLConnect ("tcp", recip_host, portnum)) {
	sprintf (line, "mailcmd internal: connecting server `%s' failed!\n",
		 recip_host);
	vxlog (0, "initialising SMTP connection failed!\n");
	deliver_error (rorig, sender, rorig, tfp, line, err);
	return;
    }
    /* E-Mail per SMTP bertragen */
    rewind (tfp);
    ifnull (strrchr (sender, '@')) {
	swd = gconc (__FILE__, __LINE__, sender, "@", mailhost(), NULL);
    } else {
	swd = gdupl (__FILE__, __LINE__, sender);
    }
    ev = smtp_send (hlc, swd, recipient, tfp, fp);
    gfree (swd); swd = NULL;
    /* Verbindung beenden */
    HLClose (hlc);
    /* Eventuell aufgetretene Fehler ausgeben */
    if (ev && ev < 8) {
	sprintf (line, "mailcmd internal: %s!\n", msgs[ev - 1]);
	vxlog (0, "SMTP protocol errors!\n");
	deliver_error (rorig, sender, rorig, tfp, line, err);
	return;
    }
    /* Evtl. eine Warnung in die LOG-Datei schreiben */
    if (ev == 8) {
	vxlog (0, "message to `%s' may have been delivered incompletely!\n",
		  recipient);
    }
}

static void deliver_nonlocal (char *rorig, char *sender, char *addr,
			      FILE *tfp, int err)
{
    if (domain_check (addr)) {
	/* Auslieferung an einen lokalen Empfnger */
	deliver_local (rorig, sender, addr, tfp, err);
    } else {
	/* Auslieferung an einen nicht-lokalen Empfnger */
	char *relayhost;
	/* Erst mal gucken, ob das berhaupt mglich ist. */
	ifnull (relayhost = prefs.nonlocalto) {
	    deliver_error (rorig, sender, addr, tfp,
			   "mail recipient not found\n", err);
	    return;
	}
	deliver_smtp (rorig, sender, relayhost, addr, tfp, err);
    }
}

/* deliver_bsmtp (sender, bsmtp_file, recipient, inputfile, msglimit, fldlimit,
**                err)
**     hngt die in `inputfile' stehende E-Mail an den Folder `bsmtp_file'
**     (VxMail-Folder) als Folge von SMTP-Kommandos (+ Daten) an.
**     Diese `bsmtp'-Variante soll eine vom SMTP-Server unabhngige Variante
**     des Mail-Relaying ermglichen; das sptere abliefern an den eigentlichen
**     Zielserver erfolgt dann durch Authentifikation mittels POP3, wobei dann
**     aber der Inhalt von `bsmtp_file' nicht per POP3, sondern per SMTP an die
**     Adresse, die aus der POP3-Verbindung hervorgeht. Der POP3-Server liefert
**     an dieser Stelle dann nur eine leere Liste von E-Mails, da er nur die
**     Authentifikation bernimmt.
*/
static void deliver_bsmtp (char *_sender, char *_batchfile,
			   char *recipient, FILE *tfp,
			   size_t ml, off_t fl, int err)
{
    xFile bFx;
    int lasteol;
    char *batchfile, line[1024], *recip, *p, *q, *sender;
    off_t sofs = skip_From (tfp);
    size_t ifsz;
    vxlog (5, "Delivering via BSMTP ...\n");
    p = strchr (_batchfile, ':');
    /* Eingefgt um 19:40 am 10.5.2000, um die Konfiguration zum Entfernen
    ** des Domainanteils aus der Empfngeradresse benutzerbezogen gestalten
    ** zu knnen.
    */
    ifnull (p) { p = strchr (_batchfile, '!'); }
    q = strchr (_batchfile, '@');
    if (p!=NULL && q!=NULL && p!=_batchfile && q!=_batchfile && q-p < 0) {
	_batchfile = p + 1;
    } else if (p == _batchfile) {
	/* Eingefgt um 19:40 am 10.5.2000, um die Konfiguration zum Entfernen
	** des Domainanteils aus der Empfngeradresse auch dann zu ermglichen,
	** wenn keine Fallbackadresse angegeben werden soll.
	*/
	_batchfile = p + 1;
    }
    batchfile = abs_path (prefs.listdir, _batchfile);
    fseek (tfp, sofs, SEEK_SET);
    /* Berechne die Lnge der neuen E-Mail nach `byte stuffing' */
    lasteol = 1; ifsz = 0;
    while (fgets (line, 1024, tfp)) {
	int eolf = cuteol (line);
	if (lasteol && *line == '.') { ifsz += 1; }
	ifsz += strlen (line);
	if (eolf) { ifsz += 2; }
	lasteol = eolf;
    }
    if (!lasteol) { ifsz += 2; }
    ifsz += 3 + strlen (_sender) + strlen (recipient) + 32;
    if (ml && ifsz > ml) {
	rewind (tfp);
	gfree (batchfile);
	vxlog (0, "maximum message size exceeded!\n");
	deliver_error (recipient, _sender, recipient, tfp,
		       "your message exceeds the allowed size", err);
	return;
    }
    fseek (tfp, sofs, SEEK_SET);
    if (!(bFx = flOpen (batchfile, "wx", prefs.fperm))) {
	vxerror_at (EX_OSERR, errno, __FILE__, __LINE__ - 1,
		    "failed to write to file `%s'!", batchfile);
    }
    xChown (bFx, geteuid (), getgid ());
    xChmod (bFx, prefs.fperm);
    if (fl && xSize(bFx)+(off_t)ifsz > fl) {
	xClose (bFx);
	rewind (tfp);
	gfree (batchfile);
	deliver_error (recipient, _sender, recipient, tfp,
		       "the mailbox is full", err);
	return;
    }
    if (strchr (_sender, '@') == NULL) {
	sender = gconc (__FILE__, __LINE__, _sender, "@", mailhost (), NULL);
    } else {
	sender = gdupl (__FILE__, __LINE__, _sender);
    }
    xPuts ("MAIL FROM:<", bFx); xPuts (sender, bFx); xPuts (">\r\n", bFx);
    vxlog (5, "MAIL FROM:<%s>\\r\\n\n", sender);
    gfree (sender);
    recip = gdupl (__FILE__, __LINE__, recipient);
    xPuts ("RCPT TO:<", bFx); xPuts (recip, bFx); xPuts (">\r\n", bFx);
    vxlog (5, "RCPT TO:<%s>\\r\\n\n", recip);
    gfree (recip);
    xPuts ("DATA\r\n", bFx);
    vxlog (5, "DATA\\r\\n\n");
    lasteol = 1;
    while (fgets (line, 1024, tfp)) {
	int eolf = cuteol (line);
	if (lasteol && *line == '.') {
	    xPuts (".", bFx);
	    vxlog (5, ".");
	}
	xPuts (line, bFx);
	vxlog (5, "%s", line);
	if (eolf) {
	    xPuts ("\r\n", bFx);
	    vxlog (5, "\\r\\n\n");
	}
	lasteol = eolf;
    }
    if (!lasteol) { xPuts ("\r\n", bFx); vxlog (5, "\\r\\n\n"); }
    xPuts (".\r\n", bFx); vxlog (5, ".\\r\\n\n");
    flSetNextUID (bFx);
    xClose (bFx);
    rewind (tfp);
    vxlog (5, "... succeeded.\n");
    gfree (batchfile);
}

static off_t skip_From (FILE *tfp)
{
    off_t fp = 0;
    char line[1024];
    rewind (tfp);
    if (fgets (line, 1024, tfp) && !strncmp (line, "From ", 5)) {
	fp += strlen (line);
	while (!cuteol (line)) {
	    if (!fgets (line, 1024, tfp)) break;
	    fp += strlen (line);
	}
    }
    return fp;
}
