/* vxdomcheck.c
**
** Domain checker - for use in conjunction with the `smail' configuration
** file `/etc/smail/routers'
**
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include "config.h"
#include "prefs.h"
#include "conf.h"
#include "dbif.h"
#include "tools.h"
#include "vxerror.h"
#include "version.h"
#include "vxopt.h"
#include "mrmacs.h"

void usage (void)
{
    fprintf (stderr, "\nUsage: %s [-f <configfile>] [-e '<successfmt>|"
		     "<failfmt>'] [-t] <hostname>"
		     "\n       %s -h"
		     "\n       %s -V\n"
		     "\n   -f <configfile>"
		     "\n      read configuration from file <configfile>"
		     "\n      (default: %s)"
		     "\n   -h"
		     "\n      show this help"
		     "\n   -V"
		     "\n      show the program's version number"
		     "\n   -e \"<successfmt>|<failfmt>\""
		     "\n      print the query results using <format string>"
		     "\n      The following `%%x' format elements will be"
		     "substituted:"
		     "\n       %%C    by the status code (OK, FAIL, ...)"
		     "\n       %%H    by the VxMail domain"
		     "\n       %%M    by the mailer name"
		     "\n       %%s    by a status message\n"
		     "\n   -t"
		     "\n      return always true - even if the database does "
		     "not exist.\n\n",
		     progname, progname, progname, CONFIGFILE);
    exit (EX_USAGE);
}

static void format(FILE *f, char *fmt, char *state, char *domain, char *mailer,
			    char *message);

int main (int argc, char *argv[])
{
    char *vxdomain, *transport, *fmt = NULL;
    DBIF_FILE userdb;
    dbif_rec *drec;
    int c, verbose = 0, always_true = 0;

    vxopt (&argc, argv, usage);
    reset_umask();

    /* Prevent error handling by `getopt' - this is MY JOB! */
    opterr = 0;

    /* Option checking: */
    for (;;) {
	c = getopt (argc, argv, "e:t");
	if (c == -1) break;
	switch (c) {
	    case 'e':
		verbose = 1; fmt = optarg; break;
	    case 't':
		always_true = 1; break;
	    default:
		vxerror_at (EX_USAGE, 0, __FILE__, __LINE__,
			    "unknown option `%s'! "
			    "Use `%s -h' for help, please!\n",
			    argv[optind], progname);
	}
    }
    if (optind == argc) {
	vxerror (EX_USAGE, 0,
		 "%s: missing argument! Use `%s -h' for help, please!\n",
		 progname, progname);
    }
    transport = "+";
    if (optind < argc - 1) { transport = argv[optind++]; }
    vxdomain = argv[optind++];
    getconf ();
    get_ug ();
    setgid (prefs.vxgid);
    setreuid (getuid(), prefs.vxuid);
    dbif_reset (&userdb, prefs.userdb_name);
    if (!userdb) {
	ifnull (fmt) {
	    printf ("ERROR invalid/missing user database.\n");
	} else {
	    format (stdout, fmt, "ERROR", NULL, NULL,
			    "invalid/missing user database");
	}
	exit (always_true ? 0 : EX_SOFTWARE);
    }
    drec = dbif_resolve (userdb, vxdomain, NULL, 0);
    dbif_close (userdb);
    if (!drec) {
	unlessnull (fmt) {
	    format (stdout, fmt, "FAIL", vxdomain, NULL, "no VxMail domain");
	}
	exit (always_true ? 0 : 1);
    }
    /* Exim: "%C %M %H + %s|%C %H - %s" */
    if (dbif_rectype(drec) != RT_DOMAIN) {
	unlessnull (fmt) {
	    format (stdout, fmt, "FORCEFAIL", vxdomain, NULL, "no domain");
	}
	exit (always_true ? 0 : EX_USAGE);
    }
    unlessnull (fmt) {
	format (stdout, fmt, "OK", vxdomain, transport, "is a VxMail domain");
    }
    exit (0);
    return 0;
}

static void ecout (FILE *f, char *p, char *q)
{
    char *r;
    int em = 0;
    for (r = q; r != p; ++r) {
	if (*r == '\\') { em = 1; continue; }
	if (*r == '_') { fputc ((em ? '_' : ' '), f); em = 0; continue; }
	if (em) { fputc ('\\', f); }
	fputc (*r, f);
	em = 0;
    }
    if (em) { fputc ('\\', f); }
}

static void format(FILE *f, char *fmt, char *state, char *domain, char *mailer,
			    char *message)
{
    char *p, *q, fc, *ov;
    p = fmt;
    q = p;
    if (strcmp (state, "OK")) {
	while (*p && *p != '|') { ++p; }
	if (!*p) { return; }
	++p; q = p;
    }
    for (;;) {
	while (*p && *p != '|' && *p != '%') { ++p; }
	ecout (f, p, q);
/*	fprintf (f, "%.*s", (p - q), q);*/
	if (!*p || *p == '|') { break; }
	++p; fc = *p++;
	switch (fc) {
	    case '%': ov = "%"; break;
	    case 'C': ov = state; break;
	    case 'H': ov = domain; break;
	    case 'M': ov = mailer; break;
	    case 's': ov = message; break;
	    default : ov = NULL; break;
	}
	unlessnull (ov) { fprintf (f, "%s", ov); }
	q = p;
    }
    fputc ('\n', f);
}
