/* uf.c
**
** file handle for the user folder
**
*/

#define tmpfile __XXX_tmpfile
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <fcntl.h>

#include "vxerror.h"
/*#include "config.h"*/
#include "tools.h"
#include "prefs.h"
#include "conv.h"
#include "exithl.h"
#include "inline-funcs.h"
#include "mrmacs.h"
#include "sigwrapper.h"

#define _UF_C
#include "uf.h"
#include "popmsg.h"
#include "bsmtp.h"
#include "vxhost.h"
#include "userdb.h"
#include "uidl.h"
#undef tmpfile

struct msg_s *msgtbl = (struct msg_s *) 0;
xFile uf_folder = (xFile) 0;
char *uf_foldername = (char *) 0;
int pop3_noopvalid = 1;

static char *tmpfile = (char *) 0;
static int tempix = 0;

/* uf_update ()
**     entfernt alle als gelscht markierten Mails aus dem Folder des
**     Benutzers - dem im folgenden als Zielfolder bezeichneten Folder.
**     Da ich hierzu `file locking' dringend bentige, darf ich die
**     ursprngliche Folderdatei nicht schlieen oder lschen; ich mu also
**     folgenden Weg gehen:
**       Ohne den Zielfolder zu schlieen, kopiere ich alle nicht als
**       gelscht markierten Mails aus diesem in einen temporren Folder.
**       Anschlieend ffne ich den temporren Folder zum lesen neu und
**       den Zielfolder mit `u'-Modus - aber mit einem weiteren Dateihandle;
**       letzteres bewirkt ein lschen smtlicher im Folder enthaltener
**       Daten.
**       Nun kopiere ich smtliche Mails aus dem temporren Folder zurck in
**       den Zielfolder.
**       Zum Schlu schliee ich das zustzliche Dateihandle auf den
**       Zielfolder, schliee den temporren Folder und lsche letzteren
**       anschlieend.
*/
void uf_update (void)
{
    xFile tmpf;
    ssize_t wlen;
    char *line[1024];
    int ix, no_update = 1;
    /* Wenn noch kein Folder geffnet wurde, dann brich hier und jetzt ab... */
    ifnull (uf_folder) { return; }
    ifnull (msgtbl) { return; }
    /* berprfe zunchst einmal, ob berhaupt Messages gelscht werden
    ** mssen.
    */
    for (ix = 0; msgtbl[ix].length >= 0; ++ix) {
	if (msgtbl[ix].deleted) { no_update = 0; }
    }
    /* Wenn nmlich nicht, dann sind die ganzen Kopieroperationen ohne einen
    ** gesteigerten Nutzen und wir ersparen sie uns!
    */
    if (no_update) {
	/* Bei temporr konvertierten Mailboxen mu die Konvertierung noch
	** rckgngig gemacht werden - indem die Indexdatei gelscht wird.
	*/
	if (tempix) { xNoIndex (uf_foldername); }
	return;
    }
    ifnull (tmpf = flOpen (tmpfile, "w", 0600)) {
	pop_msg (ERR, "i couldn't create the temporary mailbox for updating");
	return;
    }
    flCopyNextUID (uf_folder, tmpf);
    /* Kopiere zunchst alle nicht als gelscht markierten Mails
    ** in einen temporren Folder.
    */
    for (ix = 0; msgtbl[ix].length >= 0; ++ix) {
	if (msgtbl[ix].deleted) { continue; }
	xSeek (uf_folder, ix);
	while ((wlen = xRead (uf_folder, line, 1024)) > 0) {
	    xWrite (tmpf, line, (size_t) wlen);
	}
	flCopyUID (uf_folder, tmpf);
	xSet (tmpf);
    }
    /* ffne den temporren Folder neu zum lesen */
    xReSet (tmpf);
    /* Mit dem neuen Mechanismus bin ich weit besser bedient. Anstatt nmlich
    ** ein neues xFile-Handle auf den selben Folder zu generieren, leere ich
    ** diesen stattdessen einfach.
    */
    xReWrite (uf_folder);
    flCopyNextUID (tmpf, uf_folder);
    /* Kopiere nacheinander alle Mails aus dem temporren Folder in den
     * Zielfolder
     */
    for (ix = 0; xSeek (tmpf, ix) >= 0; ++ix) {
	while ((wlen = xRead (tmpf, line, 1024)) > 0) {
	    xWrite (uf_folder, line, (size_t) wlen);
	}
	flCopyUID (tmpf, uf_folder);
	xSet (uf_folder);
    }
    /* Schliee temporren und Zielfolder */
    xClose (tmpf);
    /* Entferne den temporren Folder */
    xUnlink (tmpfile);
    /* Wenn die Indexdatei nur temporr angelegt wurde, dann mu sie nun
    ** wieder entfernt werden!
    */
    if (tempix) { xNoIndex (uf_foldername); }
}

void uf_abort (void)
{
    vxlog (5, "   removing index file (if required!)\n");
    if (uf_foldername && uf_folder && tempix) {
	xNoIndex (uf_foldername);
	xClose (uf_folder);
	uf_folder = (xFile) 0;
    }
    vxlog (5, "   index file removed!\n");
}

/* Skip the `From ' line at the top of the message */
ssize_t From_Skip (xFile fx, int ix)
{
    char line[1024];
    size_t ll, res;
    xSeek (fx, ix);
    if ((ll = xGets (line, 1024, fx)) > 0) {
	if (!strncmp (line, "From ", 5)) {
	    res = ll;
	    while (line[ll - 1] != '\n' && line[ll - 1] != '\r') {
		if ((ll = xGets (line, 1024, fx)) <= 0) { break; }
		res += ll;
	    }
	    return res;
	}
    }
    xReSet (fx);
    return 0;
}

static SIGWRP(uf_bsmtpexit)

/* SIGCHLD signal handler for the `bsmtp' sub-process. */
static void uf_bsmtpexit (int sig)
{
    int chldstat;
    pid_t pid;
    pid = wait (&chldstat);
    pop3_noopvalid = 0;
}

/* Initiate the uploading of all messages in a RT_BSMTP-folder via SMTP to
** the host determined by the POP3-connection.
*/
static int uf_smtp (char *batchfolder, char *em_addr, int nd)
{
    pid_t chld;
    int cst;
    switch ((chld = fork ())) {
	case -1: /* fork failed! */
	    vxlog (0, "POP3-triggered SMTP failed: %s\n", strerror (errno));
	    return -1;
	case  0: /* child process */
	    /* clean_exithl ();
	    ** chld = setsid();
	    */
	    setpgid (getpid(), getpid());
	    /*fclose (stdin); fclose (stdout);*/
	    vxlog (5, "BSMTP: child process started!\n");
	    if (bsmtp (RADDRstr (), batchfolder, mailhost (), em_addr, nd)) {
		vxerror_at (EX_SOFTWARE, 0, __FILE__, __LINE__ - 1, "(bsmtp)");
	    }
	    vxlog (1, "Exiting `uf_smtp'.\n");
	    exit (0);
	default: /* parent process */
	    sleep (3);
	    if (waitpid (chld, &cst, WNOHANG) == -1) {
		SIGNAL (SIGCHLD, uf_bsmtpexit);
	    } else if (WIFEXITED (cst) && WEXITSTATUS(cst) == EX_TEMPFAIL) {
		/* I must add code here for signalling the already running
		** BSMTP service for this user...
		*/
		return 1;
	    } else {
		pop3_noopvalid = 0;
	    }
	    return 0;
    }
}

/* Wann mu der setup stattfinden?
** - genau dann, wenn die Authentifizierung abgeschlossen ist, d.h. nach
**   einem `APOP'-Kommando oder nach einem `PASS'-Kommando.
** Zu diesem Zeitpunkt liegt dann der Datensatz des Benutzers bereits vor.
** 1. Erweiterung (26.4.1999, 3:02:15):
**    Der hier hineingehende Datensatz kann auch ein BSMTP-Eintrag sein;
**    in diesem Fall mu eine SMTP-bertragung zu dem Host stattfinden, dessen
**    POP3-Client gerade mit `vxpop' (ebendiese) Verbindung aufgenommen hat;
**    `vxpop' mu stattdessen mit einer leeren Liste von E-Mails zurckkehren,
**    ohne, da der BSMTP-Folder irgendwie angetastet wird.
**    `vxpop' startet zu diesem Zweck einen neuen Proze, der die Aufgabe der
**    SMTP-bertragung der im Folder stehenden E-Mails im Hintergrund
**    bernimmt, whrend `vxpop' selbst den Folder nicht anrhrt und
**    stattdessen mit einer leeren Liste zurckkehrt.
*/
int uf_setup (dbif_rec *rec)
{
#if 0
    size_t fnlen = rec->flags - rec->folder, tmplen;
#endif
    size_t nmsgs = 0;
    ssize_t cxlen;
    int ix, nodomain = 0, uid_len = xuidl_len (dbif_rectype(rec) == RT_NSUSER);
    int smtp_err;
    char pidbuf[15];
    /* Check if this is only a trigger for a SMTP-upload of a BSMTP-entry. */
    if (dbif_rectype(rec) == RT_BSMTP) {
	char *fbaddr, *folder, *p, *q;
	p = strchr (rec->folder, ':');
	ifnull (p) {
	    p = strchr (rec->folder, '!');
	    unlessnull (p) { nodomain = 1; }
	}
	q = strchr (rec->folder, '@');
	if (p!=NULL && q!=NULL && p!=rec->folder && q!=rec->folder && q-p < 0) {
	    fbaddr = galloc (__FILE__, __LINE__, p - rec->folder + 1);
	    memcpy (fbaddr, rec->folder, p - rec->folder);
	    fbaddr[p - rec->folder] = '\0';
	    folder = &p[1];
	} else if (p == rec->folder) {
	    fbaddr = NULL;
	    folder = rec->folder + 1;
	} else {
	    fbaddr = NULL;
	    folder = rec->folder;
	}
	smtp_err = uf_smtp (folder, fbaddr, nodomain);
	if (smtp_err < 0) {
	    pop_msg (ERR, "attempt to spawn BSMTP failed!");
	    return 1;
	} else if (smtp_err > 0) {
	    pop_msg (ERR, "BSMTP is already running!");
	    return 1;
	}
	if (fbaddr != NULL) { gfree (fbaddr); }
	msgtbl = (struct msg_s *) galloc (__FILE__, __LINE__,
					  sizeof (struct msg_s));
	msgtbl[nmsgs].length = -1;
	vxlog (5, "   uf_smtp done, leaving uf_setup\n");
	return 0;
    }
    /* Setup the temporary mailbox name */
    sprintf (pidbuf, "%d", getpid ());
    tmpfile = gconc (__FILE__, __LINE__,
		     prefs.tmpdir, "/", progname, ".", pidbuf, NULL);
    /* setup the user's mailbox name */
    uf_foldername = abs_path (prefs.mbox_base, rec->folder);
    /* So - hier also die zustzliche Arbeit:
    **     Wenn das Flag `nonstandard' im Benutzerdatensatz gesetzt ist,
    **     dann mu der Folder konvertiert werden.
    **     Auerdem mu - da die Indexdatei nur temporr angelegt werden soll -
    **     ein Flag gesetzt werden, da es `uf_update' ermglicht, am Ende der
    **     Update-Aktionen die Indexdatei wieder zu lschen.
    */
    tempix = !prefs.syswithidx && (dbif_rectype(rec) == RT_NSUSER);
    if (tempix) {
       	if (!(uf_folder = flOpen (uf_foldername, "iX", prefs.fperm))) {
	    if (errno == ENOENT) { goto MT_FOLDER; }
	    vxlog (1, "opening non-standard mailbox `%s' failed: %s (%d,%d)\n",
		      uf_foldername, strerror (errno), geteuid(), getgid());
	    vxlog (1, " -- you should check the permissions of your "
		      "spool directory\n");
	    if (errno == EBUSY) {
		pop_msg (ERR, "you already opened this mailbox!");
		gfree (uf_foldername);
		return -1;
	    }
	    pop_msg (ERR, "opening nonstd mailbox failed!");
	    xNoIndex (uf_foldername);
	    gfree (uf_foldername);
	    return 1;
	}
	xChown (uf_folder, geteuid (), getgid ());
	xChmod (uf_folder, prefs.fperm);
	/* Setze den Puffer auf einen greren Wert, damit eine mglichst
	** schnelle Konvertierung der Datei stattfinden kann
	*/
	xSetBufSize (262144);
	vxlog (2, "preparing nonstandard mailbox\n");
	/* Folder ist im Indexmodus geffnet! Konvertiere ihn */
	conv (uf_folder);
	/* ndere den Modus des Folders von Index- auf Lesemodus */
	xReSet (uf_folder);
	vxlog (3, "register uf_abort as an exit-function!\n");
	add_exithl ("user-folder", uf_abort);
	vxlog (2, "nonstandard mailbox ready\n");
	/* Der Folder ist jetzt bereit! */
    } else if (!(uf_folder = flOpen (uf_foldername, "rX"))) {
	if (errno == ENOENT) { goto MT_FOLDER; }
	vxlog (1, "accessing mailbox `%s' failed: %s\n",
		  uf_foldername, strerror (errno));
	if (errno == EBUSY) {
	    pop_msg (ERR, "you already opened this mailbox!");
	    gfree (uf_foldername);
	    gfree (tmpfile);
	    return -1;
	}
	pop_msg (ERR, "opening user mailbox failed!");
	gfree (uf_foldername);
	gfree (tmpfile);
	return 1;
    MT_FOLDER:
	/* Der Folder konnte nicht geffnet werden. Versuche, ihn anzulegen! */
	vxlog (3, "trying to create mailbox `%s' with UID=%d, GID=%d.\n",
		  uf_foldername, getuid (), getgid ());
	if (!(uf_folder = flOpen (uf_foldername, "wx", prefs.fperm))) {
	    vxlog (1, "creating standard mailbox `%s' failed: %s\n",
		      uf_foldername, strerror (errno));
	    vxlog (1, " -- you should check the permissions of your "
		      "spool directory\n");
	    pop_msg (ERR, "creating user mailbox failed!");
	    gfree (uf_foldername);
	    gfree (tmpfile);
	    return 1;
	}
	xChmod (uf_folder, prefs.fperm);
	/* Versetze den soeben geffneten Folder in den Lesemodus zurck */
	xReSet (uf_folder);
    }
    /* Nachdem das mit der Erffnung nun geklrt ist, gehen wir zur Bestimmung
    ** der Anzahl der Mails und ihrer einzelnen Lngen ber.
    */
    for (nmsgs = 0; (cxlen = xMsgLen (uf_folder, (int) nmsgs)) >= 0; ++nmsgs);
    msgtbl = (struct msg_s *) galloc (__FILE__, __LINE__,
				      (nmsgs + 1) * sizeof (struct msg_s));
    for (ix = 0; (cxlen = xMsgLen (uf_folder, ix)) >= 0; ++ix) {
	msgtbl[ix].length = ((ssize_t) cxlen) - From_Skip (uf_folder, ix) + uid_len;
	msgtbl[ix].deleted = 0;
    }
    msgtbl[nmsgs].length = -1;
    vxlog (5, "   leaving uf_setup\n");
    return 0;
}
