/* digest.h
**
** Interface file for the message `digest' module.
** (Collection of message digest algorithms.)
**
*/
#ifndef _DIGEST_H
#define _DIGEST_H

/* In order make this interface independent from a special message digest
** algorithm, the structure of the digest context - which is required during
** the calculation of the message digest - is hidden; this makes the digest
** module an abstract data type.
*/
struct _DIG_CTX;
typedef struct _DIG_CTX *DIG_CTX;

/* Generate a message digest context structure of the specified type.
** The type is specified as a (case independent) string.
*/
DIG_CTX digest_new (char *dgtype);

/* Release the memory allocated for the message digest context structure. */
void digest_free (DIG_CTX ctx);

/* (re-)initialize the message digest context structure. */
int digest_init (DIG_CTX ctx);

/* update the message digest context dtructure with the data in msg (of the
** length msglen).
*/
int digest_update (DIG_CTX ctx, unsigned char *msg, size_t msglen);

/* finalize the calculation of the message digest and return it in the
** argument `buf' - which should be at least 20 bytes long.
*/
int digest_final (unsigned char buf[], DIG_CTX ctx);

/* change the type of the message digest to be calculated. */
int digest_chgtype (DIG_CTX ctx, char *newtype);

/* return the size of buffer which must be allocated for the digest. */
ssize_t digest_size (DIG_CTX ctx);

#endif /* defined(_DIGEST_H) */
